from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Any, Tuple

import torch
import torch.fft
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor, nn

from nnts.utils import Scheduler, TrainingMethod

from ..datasets import PaddedData


@dataclass
class Hyperparams:
    """Class for keeping track of training and model params."""

    freq: str
    optimizer: callable = torch.optim.Adam
    loss_fn: callable = F.l1_loss
    dropout: float = 0.1
    batch_size: int = 32
    lr: float = 0.005
    epochs: int = 100
    patience: int = 10
    early_stopper_patience: int = 30
    batches_per_epoch: int = 50
    weight_decay: float = 0.0
    training_method: TrainingMethod = TrainingMethod.DMS
    scheduler: Scheduler = Scheduler.REDUCE_LR_ON_PLATEAU
    model_file_path: str = f"logs"
    top_k: int = 5
    num_kernels: int = 6
    d_model: int = 32
    embed: str = "timeF"
    d_ff: int = 32
    e_layers: int = 2
    enc_in: int = 1
    c_out: int = 1
    label_len: int = 0
    task_name: str = "short_term_forecast"


def get_mutlivariate_params():
    params = Hyperparams(
        freq="h",
        optimizer=torch.optim.Adam,
        loss_fn=torch.nn.MSELoss(),
        batch_size=32,
        batches_per_epoch=250,
        training_method=TrainingMethod.DMS,
        model_file_path="logs",
        epochs=10,
        scheduler=Scheduler.STEP_LR,
        lr=0.001,
        weight_decay=0.0,
        c_out=7,
        enc_in=7,
        d_model=16,
    )
    return params


class PositionalEmbedding(nn.Module):
    def __init__(self, d_model, max_len=5000):
        super(PositionalEmbedding, self).__init__()
        # Compute the positional encodings once in log space.
        pe = torch.zeros(max_len, d_model).float()
        pe.require_grad = False

        position = torch.arange(0, max_len).float().unsqueeze(1)
        div_term = (
            torch.arange(0, d_model, 2).float() * -(math.log(10000.0) / d_model)
        ).exp()

        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)

        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def forward(self, x):
        return self.pe[:, : x.size(1)]


class TokenEmbedding(nn.Module):
    def __init__(self, c_in, d_model):
        super(TokenEmbedding, self).__init__()
        padding = 1 if torch.__version__ >= "1.5.0" else 2
        self.tokenConv = nn.Conv1d(
            in_channels=c_in,
            out_channels=d_model,
            kernel_size=3,
            padding=padding,
            padding_mode="circular",
            bias=False,
        )
        for m in self.modules():
            if isinstance(m, nn.Conv1d):
                nn.init.kaiming_normal_(
                    m.weight, mode="fan_in", nonlinearity="leaky_relu"
                )

    def forward(self, x):
        x = self.tokenConv(x.permute(0, 2, 1)).transpose(1, 2)
        return x


class FixedEmbedding(nn.Module):
    def __init__(self, c_in, d_model):
        super(FixedEmbedding, self).__init__()

        w = torch.zeros(c_in, d_model).float()
        w.require_grad = False

        position = torch.arange(0, c_in).float().unsqueeze(1)
        div_term = (
            torch.arange(0, d_model, 2).float() * -(math.log(10000.0) / d_model)
        ).exp()

        w[:, 0::2] = torch.sin(position * div_term)
        w[:, 1::2] = torch.cos(position * div_term)

        self.emb = nn.Embedding(c_in, d_model)
        self.emb.weight = nn.Parameter(w, requires_grad=False)

    def forward(self, x):
        return self.emb(x).detach()


class TemporalEmbedding(nn.Module):
    def __init__(self, d_model, embed_type="fixed", freq="h"):
        super(TemporalEmbedding, self).__init__()

        minute_size = 4
        hour_size = 24
        weekday_size = 7
        day_size = 32
        month_size = 13

        Embed = FixedEmbedding if embed_type == "fixed" else nn.Embedding
        if freq == "t":
            self.minute_embed = Embed(minute_size, d_model)
        self.hour_embed = Embed(hour_size, d_model)
        self.weekday_embed = Embed(weekday_size, d_model)
        self.day_embed = Embed(day_size, d_model)
        self.month_embed = Embed(month_size, d_model)

    def forward(self, x):
        x = x.long()
        minute_x = (
            self.minute_embed(x[:, :, 4]) if hasattr(self, "minute_embed") else 0.0
        )
        hour_x = self.hour_embed(x[:, :, 3])
        weekday_x = self.weekday_embed(x[:, :, 2])
        day_x = self.day_embed(x[:, :, 1])
        month_x = self.month_embed(x[:, :, 0])

        return hour_x + weekday_x + day_x + month_x + minute_x


class TimeFeatureEmbedding(nn.Module):
    def __init__(self, d_model, embed_type="timeF", freq="h"):
        super(TimeFeatureEmbedding, self).__init__()

        freq_map = {"h": 4, "t": 5, "s": 6, "m": 1, "a": 1, "w": 2, "d": 3, "b": 3}
        d_inp = freq_map.get(freq, 4)
        self.embed = nn.Linear(d_inp, d_model, bias=False)

    def forward(self, x):
        return self.embed(x)


class DataEmbedding(nn.Module):
    def __init__(self, c_in, d_model, embed_type="fixed", freq="h", dropout=0.1):
        super(DataEmbedding, self).__init__()

        self.value_embedding = TokenEmbedding(c_in=c_in, d_model=d_model)
        self.position_embedding = PositionalEmbedding(d_model=d_model)
        self.temporal_embedding = (
            TemporalEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
            if embed_type != "timeF"
            else TimeFeatureEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
        )
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x, x_mark):
        if x_mark is None:
            x = self.value_embedding(x) + self.position_embedding(x)
        else:
            x = (
                self.value_embedding(x)
                + self.temporal_embedding(x_mark)
                + self.position_embedding(x)
            )
        return self.dropout(x)


def FFT_for_Period(x, k=2):
    # [B, T, C]
    xf = torch.fft.rfft(x, dim=1)
    # find period by amplitudes
    frequency_list = abs(xf).mean(0).mean(-1)
    frequency_list[0] = 0
    _, top_list = torch.topk(frequency_list, k)
    top_list = top_list.detach().cpu().numpy()
    period = x.shape[1] // top_list
    return period, abs(xf).mean(-1)[:, top_list]


class Inception_Block_V1(nn.Module):
    def __init__(self, in_channels, out_channels, num_kernels=6, init_weight=True):
        super(Inception_Block_V1, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_kernels = num_kernels
        kernels = []
        for i in range(self.num_kernels):
            kernels.append(
                nn.Conv2d(in_channels, out_channels, kernel_size=2 * i + 1, padding=i)
            )
        self.kernels = nn.ModuleList(kernels)
        if init_weight:
            self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward(self, x):
        res_list = []
        for i in range(self.num_kernels):
            res_list.append(self.kernels[i](x))
        res = torch.stack(res_list, dim=-1).mean(-1)
        return res


class TimesBlock(nn.Module):
    def __init__(
        self, context_length: int, prediction_length: int, configs: Hyperparams
    ):
        super(TimesBlock, self).__init__()
        self.seq_len = context_length
        self.pred_len = prediction_length
        self.k = configs.top_k
        # parameter-efficient design
        self.conv = nn.Sequential(
            Inception_Block_V1(
                configs.d_model, configs.d_ff, num_kernels=configs.num_kernels
            ),
            nn.GELU(),
            Inception_Block_V1(
                configs.d_ff, configs.d_model, num_kernels=configs.num_kernels
            ),
        )

    def forward(self, x):
        B, T, N = x.size()
        period_list, period_weight = FFT_for_Period(x, self.k)

        res = []
        for i in range(self.k):
            period = period_list[i]
            # padding
            if (self.seq_len + self.pred_len) % period != 0:
                length = (((self.seq_len + self.pred_len) // period) + 1) * period
                padding = torch.zeros(
                    [x.shape[0], (length - (self.seq_len + self.pred_len)), x.shape[2]]
                ).to(x.device)
                out = torch.cat([x, padding], dim=1)
            else:
                length = self.seq_len + self.pred_len
                out = x
            # reshape
            out = (
                out.reshape(B, length // period, period, N)
                .permute(0, 3, 1, 2)
                .contiguous()
            )
            # 2D conv: from 1d Variation to 2d Variation
            out = self.conv(out)
            # reshape back
            out = out.permute(0, 2, 3, 1).reshape(B, -1, N)
            res.append(out[:, : (self.seq_len + self.pred_len), :])
        res = torch.stack(res, dim=-1)
        # adaptive aggregation
        period_weight = F.softmax(period_weight, dim=1)
        period_weight = period_weight.unsqueeze(1).unsqueeze(1).repeat(1, T, N, 1)
        res = torch.sum(res * period_weight, -1)
        # residual connection
        res = res + x
        return res


class TimesNet(nn.Module):
    """
    Paper link: https://openreview.net/pdf?id=ju_Uqw384Oq
    """

    def __init__(
        self, context_length: int, prediction_length: int, configs: Hyperparams
    ):
        super(TimesNet, self).__init__()
        self.configs = configs
        self.task_name = configs.task_name
        self.seq_len = context_length
        self.label_len = configs.label_len
        self.pred_len = prediction_length
        self.model = nn.ModuleList(
            [
                TimesBlock(context_length, prediction_length, configs)
                for _ in range(configs.e_layers)
            ]
        )
        self.enc_embedding = DataEmbedding(
            configs.enc_in,
            configs.d_model,
            configs.embed,
            configs.freq,
            configs.dropout,
        )
        self.layer = configs.e_layers
        self.layer_norm = nn.LayerNorm(configs.d_model)
        self.predict_linear = nn.Linear(self.seq_len, self.pred_len + self.seq_len)
        self.projection = nn.Linear(configs.d_model, configs.c_out, bias=True)

    def forecast(self, x_enc, x_mark_enc):
        # Normalization from Non-stationary Transformer
        means = x_enc.mean(1, keepdim=True).detach()
        x_enc = x_enc - means
        stdev = torch.sqrt(torch.var(x_enc, dim=1, keepdim=True, unbiased=False) + 1e-5)
        x_enc /= stdev

        # embedding
        enc_out = self.enc_embedding(x_enc, x_mark_enc)  # [B,T,C]
        enc_out = self.predict_linear(enc_out.permute(0, 2, 1)).permute(
            0, 2, 1
        )  # align temporal dimension
        # TimesNet
        for i in range(self.layer):
            enc_out = self.layer_norm(self.model[i](enc_out))
        # porject back
        dec_out = self.projection(enc_out)

        # De-Normalization from Non-stationary Transformer
        dec_out = dec_out * (
            stdev[:, 0, :].unsqueeze(1).repeat(1, self.pred_len + self.seq_len, 1)
        )
        dec_out = dec_out + (
            means[:, 0, :].unsqueeze(1).repeat(1, self.pred_len + self.seq_len, 1)
        )
        return dec_out

    def imputation(self, x_enc, x_mark_enc, x_dec, x_mark_dec, mask):
        # Normalization from Non-stationary Transformer
        means = torch.sum(x_enc, dim=1) / torch.sum(mask == 1, dim=1)
        means = means.unsqueeze(1).detach()
        x_enc = x_enc - means
        x_enc = x_enc.masked_fill(mask == 0, 0)
        stdev = torch.sqrt(
            torch.sum(x_enc * x_enc, dim=1) / torch.sum(mask == 1, dim=1) + 1e-5
        )
        stdev = stdev.unsqueeze(1).detach()
        x_enc /= stdev

        # embedding
        enc_out = self.enc_embedding(x_enc, x_mark_enc)  # [B,T,C]
        # TimesNet
        for i in range(self.layer):
            enc_out = self.layer_norm(self.model[i](enc_out))
        # porject back
        dec_out = self.projection(enc_out)

        # De-Normalization from Non-stationary Transformer
        dec_out = dec_out * (
            stdev[:, 0, :].unsqueeze(1).repeat(1, self.pred_len + self.seq_len, 1)
        )
        dec_out = dec_out + (
            means[:, 0, :].unsqueeze(1).repeat(1, self.pred_len + self.seq_len, 1)
        )
        return dec_out

    def anomaly_detection(self, x_enc):
        # Normalization from Non-stationary Transformer
        means = x_enc.mean(1, keepdim=True).detach()
        x_enc = x_enc - means
        stdev = torch.sqrt(torch.var(x_enc, dim=1, keepdim=True, unbiased=False) + 1e-5)
        x_enc /= stdev

        # embedding
        enc_out = self.enc_embedding(x_enc, None)  # [B,T,C]
        # TimesNet
        for i in range(self.layer):
            enc_out = self.layer_norm(self.model[i](enc_out))
        # porject back
        dec_out = self.projection(enc_out)

        # De-Normalization from Non-stationary Transformer
        dec_out = dec_out * (
            stdev[:, 0, :].unsqueeze(1).repeat(1, self.pred_len + self.seq_len, 1)
        )
        dec_out = dec_out + (
            means[:, 0, :].unsqueeze(1).repeat(1, self.pred_len + self.seq_len, 1)
        )
        return dec_out

    def classification(self, x_enc, x_mark_enc):
        # embedding
        enc_out = self.enc_embedding(x_enc, None)  # [B,T,C]
        # TimesNet
        for i in range(self.layer):
            enc_out = self.layer_norm(self.model[i](enc_out))

        # Output
        # the output transformer encoder/decoder embeddings don't include non-linearity
        output = self.act(enc_out)
        output = self.dropout(output)
        # zero-out padding embeddings
        output = output * x_mark_enc.unsqueeze(-1)
        # (batch_size, seq_length * d_model)
        output = output.reshape(output.shape[0], -1)
        output = self.projection(output)  # (batch_size, num_classes)
        return output

    def forward(self, x_enc, x_mark_enc):
        if (
            self.task_name == "long_term_forecast"
            or self.task_name == "short_term_forecast"
        ):
            dec_out = self.forecast(x_enc, x_mark_enc)
            return dec_out[:, -self.pred_len :, :]  # [B, L, D]
        return None

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data[:, : self.seq_len, :]
        pad_mask = batch.pad_mask[:, : self.seq_len]
        y_hat = self(x, None)
        y = batch.data[:, self.seq_len :, :]

        pad_mask = batch.pad_mask[:, self.seq_len :]
        y_hat = y_hat[pad_mask]
        y = y[pad_mask]

        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        y_hat = self(x, None)
        return y_hat



# This source code is provided for the purposes of scientific reproducibility
# under the following limited license from Element AI Inc. The code is an
# implementation of the N-BEATS model (Oreshkin et al., N-BEATS: Neural basis
# expansion analysis for interpretable time series forecasting,
# https://arxiv.org/abs/1905.10437). The copyright to the source code is
# licensed under the Creative Commons - Attribution-NonCommercial 4.0
# International license (CC BY-NC 4.0):
# https://creativecommons.org/licenses/by-nc/4.0/.  Any commercial use (whether
# for the benefit of third parties or internally in production) requires an
# explicit license. The subject-matter of the N-BEATS model and associated
# materials are the property of Element AI Inc. and may be subject to patent
# protection. No license to patents is granted hereunder (whether express or
# implied). Copyright © 2020 Element AI Inc. All rights reserved.

"""
N-BEATS Model.
"""
from dataclasses import dataclass
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from nnts.utils import Scheduler, TrainingMethod

from ..datasets import PaddedData


@dataclass
class Hyperparams:
    """Class for keeping track of training and model params."""

    optimizer: callable = torch.optim.Adam
    loss_fn: callable = F.l1_loss
    dropout: float = 0.0

    lr: float = 1e-3
    epochs: int = 100
    patience: int = 10
    early_stopper_patience: int = 30
    batches_per_epoch: int = 50
    weight_decay: float = 0.0
    training_method: TrainingMethod = TrainingMethod.DMS
    scheduler: Scheduler = Scheduler.REDUCE_LR_ON_PLATEAU
    model_file_path = f"logs"

    # Paper
    # batch_size: int = 1024

    # GluonTS
    theta_size: int = 32
    num_stacks: int = 30
    num_layers: int = 2
    layer_size: int = 512
    batch_size: int = 32

    # Darts
    # theta_size: int = 5
    # num_stacks: int = 30
    # num_layers: int = 4
    # layer_size: int = 256


class GenericBasis(nn.Module):
    """
    Generic basis function.
    """

    def __init__(self, theta_size: int):
        super().__init__()
        self.theta_size = theta_size

    def forward(self, theta: torch.Tensor):
        return theta[:, : self.theta_size], theta[:, -self.theta_size :]


class TrendBasis(nn.Module):
    """
    Polynomial function to model trend.
    """

    def __init__(
        self, degree_of_polynomial: int, backcast_size: int, forecast_size: int
    ):
        super().__init__()
        self.polynomial_size = (
            degree_of_polynomial + 1
        )  # degree of polynomial with constant term

        # Create the backcast time tensor using PyTorch methods
        backcast_time = torch.cat(
            [
                (torch.arange(backcast_size, dtype=torch.float32) / backcast_size)
                .pow(i)
                .unsqueeze(0)
                for i in range(self.polynomial_size)
            ]
        )
        self.backcast_time = nn.Parameter(backcast_time, requires_grad=False)

        # Create the forecast time tensor using PyTorch methods
        forecast_time = torch.cat(
            [
                (torch.arange(forecast_size, dtype=torch.float32) / forecast_size)
                .pow(i)
                .unsqueeze(0)
                for i in range(self.polynomial_size)
            ]
        )
        self.forecast_time = nn.Parameter(forecast_time, requires_grad=False)

    def forward(self, theta: torch.Tensor):
        # Compute the backcast and forecast using einsum for matrix multiplication
        backcast = torch.einsum(
            "bp,pt->bt", theta[:, self.polynomial_size :], self.backcast_time
        )
        forecast = torch.einsum(
            "bp,pt->bt", theta[:, : self.polynomial_size], self.forecast_time
        )
        return backcast, forecast


class SeasonalityBasis(nn.Module):
    """
    Harmonic functions to model seasonality.
    """

    def __init__(self, harmonics: int, backcast_size: int, forecast_size: int):
        super().__init__()

        # Define frequency tensor
        frequency = torch.cat(
            (
                torch.zeros(1, dtype=torch.float32),
                torch.arange(
                    harmonics, harmonics / 2 * forecast_size, dtype=torch.float32
                )
                / harmonics,
            )
        ).unsqueeze(
            0
        )  # Shape: (1, num_frequencies)

        # Define backcast and forecast grids
        backcast_grid = (
            -2
            * torch.pi
            * (
                torch.arange(backcast_size, dtype=torch.float32).unsqueeze(1)
                / forecast_size
            )
            * frequency
        )
        forecast_grid = (
            2
            * torch.pi
            * (
                torch.arange(forecast_size, dtype=torch.float32).unsqueeze(1)
                / forecast_size
            )
            * frequency
        )

        # Create cos and sin templates for backcast and forecast using PyTorch
        self.backcast_cos_template = nn.Parameter(
            torch.cos(backcast_grid).transpose(0, 1), requires_grad=False
        )
        self.backcast_sin_template = nn.Parameter(
            torch.sin(backcast_grid).transpose(0, 1), requires_grad=False
        )
        self.forecast_cos_template = nn.Parameter(
            torch.cos(forecast_grid).transpose(0, 1), requires_grad=False
        )
        self.forecast_sin_template = nn.Parameter(
            torch.sin(forecast_grid).transpose(0, 1), requires_grad=False
        )

    def forward(self, theta: torch.Tensor):
        # Calculate the number of parameters per harmonic
        params_per_harmonic = theta.shape[1] // 4

        # Compute backcast harmonics using einsum for efficient batched multiplication
        backcast_harmonics_cos = torch.einsum(
            "bp,pt->bt",
            theta[:, 2 * params_per_harmonic : 3 * params_per_harmonic],
            self.backcast_cos_template,
        )
        backcast_harmonics_sin = torch.einsum(
            "bp,pt->bt", theta[:, 3 * params_per_harmonic :], self.backcast_sin_template
        )
        backcast = backcast_harmonics_cos + backcast_harmonics_sin

        # Compute forecast harmonics using einsum
        forecast_harmonics_cos = torch.einsum(
            "bp,pt->bt", theta[:, :params_per_harmonic], self.forecast_cos_template
        )
        forecast_harmonics_sin = torch.einsum(
            "bp,pt->bt",
            theta[:, params_per_harmonic : 2 * params_per_harmonic],
            self.forecast_sin_template,
        )
        forecast = forecast_harmonics_cos + forecast_harmonics_sin

        return backcast, forecast


class NBeatsBlock(nn.Module):
    """
    N-BEATS block which takes a basis function as an argument.
    """

    def __init__(
        self,
        h: int,
        input_size,
        theta_size: int,
        basis_function: nn.Module,
        layers: int,
        layer_size: int,
    ):
        """
        N-BEATS block.

        :param input_size: Insample size.
        :param theta_size:  Number of parameters for the basis function.
        :param basis_function: Basis function which takes the parameters and produces backcast and forecast.
        :param layers: Number of layers.
        :param layer_size: Layer size.
        """
        super().__init__()
        self.layers = nn.ModuleList(
            [nn.Linear(in_features=input_size, out_features=layer_size)]
            + [
                nn.Linear(in_features=layer_size, out_features=layer_size)
                for _ in range(layers - 1)
            ]
        )
        self.basis_parameters = nn.Linear(
            in_features=layer_size, out_features=theta_size * 2
        )
        self.basis_function = basis_function
        self.backcast = nn.Linear(in_features=theta_size, out_features=input_size)
        self.forecast = nn.Linear(in_features=theta_size, out_features=h)

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        block_input = x
        for layer in self.layers:
            block_input = F.relu(layer(block_input))
        basis_parameters = self.basis_parameters(block_input)
        backcast, forecast = self.basis_function(basis_parameters)
        return self.backcast(backcast), self.forecast(forecast)


class NBeats(nn.Module):
    """
    N-Beats Model.
    """

    def __init__(self, h: int, input_size: int, configs: Hyperparams):
        super().__init__()
        self.blocks = nn.ModuleList(
            [
                NBeatsBlock(
                    h=h,
                    input_size=input_size,
                    theta_size=configs.theta_size,
                    basis_function=GenericBasis(configs.theta_size),
                    layers=configs.num_layers,
                    layer_size=configs.layer_size,
                )
                for _ in range(configs.num_stacks)
            ]
        )
        self.seq_len = input_size
        self.h = h

    def forward(self, x: torch.Tensor, input_mask: torch.Tensor) -> torch.Tensor:
        x = x.squeeze(-1)
        input_mask = input_mask.squeeze(-1)
        residuals = x.flip(dims=(1,))
        input_mask = input_mask.flip(dims=(1,))
        forecast = x[:, -1:]
        for i, block in enumerate(self.blocks):
            backcast, block_forecast = block(residuals)
            residuals = (residuals - backcast) * input_mask
            forecast = forecast + block_forecast
        return forecast.unsqueeze(-1)

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data[:, : self.seq_len, :]
        x_pad_mask = batch.pad_mask[:, : self.seq_len]
        y_hat = self(x, x_pad_mask)
        y = batch.data[:, self.seq_len :, :]

        pad_mask = batch.pad_mask[:, self.seq_len :]
        y_hat = y_hat[pad_mask]
        y = y[pad_mask]

        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        y_hat = self(x, pad_mask)
        return y_hat



from typing import Dict, Tuple

import torch
import torch.nn as nn

from nnts import utils

from .. import datasets


class UnrolledLSTMDecoder(nn.Module):
    def __init__(self, params: utils.Hyperparams, output_dim: int):
        super(UnrolledLSTMDecoder, self).__init__()
        self.params = params

        if params.rnn_type == "gru":
            self.rnn = nn.GRU(
                output_dim + 1,
                params.hidden_dim,
                params.n_layers,
                dropout=params.dropout,
                batch_first=True,
            )
        else:
            self.rnn = nn.LSTM(
                output_dim + 1,
                params.hidden_dim,
                params.n_layers,
                dropout=params.dropout,
                batch_first=True,
            )
        self.rnn_type = params.rnn_type

    def init_hidden_zeros(self, batch_size: int):
        if self.rnn_type == "gru":
            hidden = torch.zeros(
                self.params.n_layers, batch_size, self.params.hidden_dim
            )
        else:
            h0 = torch.zeros(self.params.n_layers, batch_size, self.params.hidden_dim)
            c0 = torch.zeros(self.params.n_layers, batch_size, self.params.hidden_dim)
            hidden = (h0, c0)
        return hidden

    def forward(self, X: torch.tensor, hidden=None):
        """
        H: number of steps to unroll
        """
        B, T, C = X.shape
        if hidden is None:
            hidden = self.init_hidden_zeros(B)
        out, hidden = self.rnn(X, hidden)
        return out, hidden


class UnrolledLSTM(nn.Module):

    def __init__(
        self,
        Distribution: nn.Module,
        params: utils.Hyperparams,
        scaling_fn: callable,
        output_dim: int,
    ):
        super(UnrolledLSTM, self).__init__()
        self.scaling_fn = scaling_fn
        self.decoder = UnrolledLSTMDecoder(params, output_dim)
        self.distribution = Distribution(params.hidden_dim, output_dim)

    def forward(self, X: torch.tensor, pad_mask: torch.tensor, H: int) -> torch.tensor:
        X = X.clone()
        B, T, C = X.shape
        y_hat = torch.zeros(B, T + H, C)

        target_scale = self.scaling_fn(X, pad_mask)
        conts = X / target_scale
        embedded = torch.cat(
            [conts, torch.log(target_scale[:, :, :1]).expand(B, T, 1)], 2
        )
        out, hidden = self.decoder(embedded)
        out = self.distribution(out, target_scale=None)
        y_hat[:, :T, :] = out
        for t in range(0, H):
            embedded = torch.cat([out[:, -1:, :], torch.log(target_scale[:, :, :1])], 2)
            out, hidden = self.decoder(embedded, hidden)
            out = self.distribution(out, target_scale=None)
            y_hat[:, T + t, :] = out[:, -1, :]

        y_hat = y_hat * target_scale
        return y_hat

    def generate(
        self,
        X: torch.tensor,
        pad_mask: torch.tensor,
        prediction_length: int,
        context_length: int,
    ) -> torch.tensor:
        y_hat = self.forward(
            X[:, -context_length:, :], pad_mask[:, -context_length:], prediction_length
        )
        y_hat = y_hat[:, -prediction_length:, :]
        return y_hat

    def free_running(
        self, batch: datasets.PaddedData, context_length: int, prediction_length: int
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data
        pad_mask = batch.pad_mask

        y = x[:, 1:, :]

        y_hat = self.forward(
            x[:, :context_length, :],
            pad_mask[:, :context_length],
            prediction_length - 1,
        )
        y_hat = y_hat[pad_mask[:, 1:]]
        y = y[pad_mask[:, 1:]]
        return y_hat, y

    def teacher_forcing_output(
        self, batch: datasets.PaddedData
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data
        pad_mask = batch.pad_mask
        y_hat = self(x[:, :-1, :], pad_mask[:, :-1], H=0)
        y = x[:, 1:, :]
        y_hat = y_hat[pad_mask[:, 1:]]
        y = y[pad_mask[:, 1:]]
        return y_hat, y



import json
import os
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Callable

import pandas as pd


class TrainingMethod(Enum):
    TEACHER_FORCING = auto()
    FREE_RUNNING = auto()
    DMS = auto()


class Scheduler(Enum):
    ONE_CYCLE = auto()
    REDUCE_LR_ON_PLATEAU = auto()
    STEP_LR = auto()


@dataclass
class Hyperparams:
    """Class for keeping track of training and model params."""

    optimizer: Any
    loss_fn: Any
    input_dim: int = 1
    hidden_dim: int = 40
    n_layers: int = 2
    dropout: float = 0.1
    output_dim: int = 1
    batch_size: int = 128
    lr: float = 1e-3
    epochs: int = 100
    patience: int = 10
    rnn_type: str = "lstm"
    early_stopper_patience: int = 30
    batches_per_epoch: int = 200
    weight_decay: float = 1e-8
    training_method: TrainingMethod = TrainingMethod.TEACHER_FORCING
    scheduler: Scheduler = Scheduler.ONE_CYCLE
    model_file_path: str = ""


@dataclass
class GluonTsDefaultWithOneCycle(Hyperparams):
    """Class for keeping track of training and model params."""

    optimizer: Callable
    loss_fn: Callable
    input_dim: int = 1
    hidden_dim: int = 40
    n_layers: int = 2
    dropout: float = 0.1
    output_dim: int = 1
    batch_size: int = 32
    lr: float = 1e-3
    epochs: int = 100
    patience: int = 10
    rnn_type: str = "lstm"
    early_stopper_patience: int = 30
    batches_per_epoch: int = 50
    weight_decay: float = 1e-8
    training_method: TrainingMethod = TrainingMethod.TEACHER_FORCING
    scheduler: Scheduler = Scheduler.REDUCE_LR_ON_PLATEAU


def makedirs_if_not_exists(directory: str) -> None:
    if len(directory) == 0:
        return
    if not os.path.exists(directory):
        os.makedirs(directory)


class CSVFileAggregator:
    """
    A class to aggregate JSON files in a directory and save them as a CSV file.

    Parameters
    ----------
    path : str
        The directory path where JSON files are located.
    filename : str
        The name of the output CSV file (without extension).

    Methods
    -------
    __call__() -> pd.DataFrame
        Aggregates JSON files in the specified directory, saves the data as a CSV file,
        and returns the aggregated data as a DataFrame.
    """

    def __init__(self, path: str, filename: str):
        """
        Initializes the CSVFileAggregator with the given directory path and output filename.

        Parameters
        ----------
        path : str
            The directory path where JSON files are located.
        filename : str
            The name of the output CSV file (without extension).
        """
        self.path = path
        self.filename = filename

    def __call__(self) -> pd.DataFrame:
        """
        Aggregates JSON files in the specified directory, saves the data as a CSV file,
        and returns the aggregated data as a DataFrame.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the aggregated data from the JSON files.

        Raises
        ------
        FileNotFoundError
            If the specified path does not exist.
        """
        data_list = []
        for filename in os.listdir(self.path):
            if filename.endswith(".json"):
                with open(os.path.join(self.path, filename), "r") as file:
                    data = json.load(file)
                    data_list.append(data)
        # Concatenate DataFrames if needed
        results = pd.DataFrame(data_list)
        results.to_csv(f"{self.path}/{self.filename}.csv", index=False)
        return results



from typing import Dict, Tuple

import torch
import torch.nn as nn

from nnts import utils

from .. import datasets


class LinearModel(nn.Module):
    """
    This model predicts a point predction for ordinal data.
    """

    def __init__(self, hidden_size: int, output_size: int):
        super(LinearModel, self).__init__()
        self.main = nn.Sequential(
            nn.Linear(hidden_size, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, output_size),
        )

    def forward(self, x: torch.tensor, target_scale: torch.tensor):
        y_hat = self.main(x)
        if target_scale is not None:
            y_hat = y_hat * target_scale
        return y_hat


class BaseLSTMDecoder(nn.Module):
    def __init__(self, params: utils.Hyperparams, output_dim: int):
        super(BaseLSTMDecoder, self).__init__()
        self.params = params

        if params.rnn_type == "gru":
            self.rnn = nn.GRU(
                output_dim + 1,
                params.hidden_dim,
                params.n_layers,
                dropout=params.dropout,
                batch_first=True,
            )
        else:
            self.rnn = nn.LSTM(
                output_dim + 1,
                params.hidden_dim,
                params.n_layers,
                dropout=params.dropout,
                batch_first=True,
            )
        self.rnn_type = params.rnn_type

    def init_hidden_zeros(self, batch_size: int):
        if self.rnn_type == "gru":
            hidden = torch.zeros(
                self.params.n_layers, batch_size, self.params.hidden_dim
            )
        else:
            h0 = torch.zeros(self.params.n_layers, batch_size, self.params.hidden_dim)
            c0 = torch.zeros(self.params.n_layers, batch_size, self.params.hidden_dim)
            hidden = (h0, c0)
        return hidden

    def forward(self, X: torch.tensor, hidden=None):
        B, T, C = X.shape
        if hidden is None:
            hidden = self.init_hidden_zeros(B)
        return self.rnn(X, hidden)


class BaseLSTM(nn.Module):

    def __init__(
        self,
        Distribution: nn.Module,
        params: utils.Hyperparams,
        scaling_fn: callable,
        output_dim: int,
    ):
        super(BaseLSTM, self).__init__()
        self.scaling_fn = scaling_fn
        self.decoder = BaseLSTMDecoder(params, output_dim)
        self.distribution = Distribution(params.hidden_dim, output_dim)

    def forward(self, X: torch.tensor, pad_mask: torch.tensor) -> torch.tensor:
        X = X.clone()
        B, T, C = X.shape
        target_scale = self.scaling_fn(X, pad_mask)
        conts = X / target_scale
        embedded = torch.cat(
            [conts, torch.log(target_scale[:, :, :1]).expand(B, T, 1)], 2
        )
        out, _ = self.decoder(embedded)
        distr = self.distribution(out, target_scale=target_scale)
        return distr

    def generate(
        self,
        X: torch.tensor,
        pad_mask: torch.tensor,
        prediction_length: int,
        context_length: int,
    ) -> torch.tensor:
        pred_list = []
        while True:
            pad_mask = pad_mask[:, -context_length:]
            assert (
                pad_mask.sum()
                == X[:, -context_length:, :].shape[0]
                * X[:, -context_length:, :].shape[1]
            )
            preds = self.forward(X[:, -context_length:, :], pad_mask)
            # focus only on the last time step
            preds = preds[:, -1:, :]  # becomes (B, 1, C)
            pred_list.append(preds)

            if len(pred_list) >= prediction_length:
                break
            y_hat = preds.detach().clone()
            X = torch.cat((X, y_hat), dim=1)  # (B, T+1)
            pad_mask = torch.cat((pad_mask, torch.ones_like(preds[:, :, 0])), dim=1)
        return torch.cat(pred_list, 1)

    def teacher_forcing_output(self, batch: datasets.PaddedData, *args, **kwargs):
        x = batch.data
        pad_mask = batch.pad_mask
        y_hat = self(x[:, :-1, :], pad_mask[:, :-1])
        y = x[:, 1:, :]

        y_hat = y_hat[pad_mask[:, 1:]]
        y = y[pad_mask[:, 1:]]
        return y_hat, y


class BaseFutureCovariateLSTM(nn.Module):

    def __init__(
        self,
        Distribution: nn.Module,
        params: utils.Hyperparams,
        scaling_fn: callable,
        output_dim: int,
        known_future_covariates: int,
    ):
        super(BaseFutureCovariateLSTM, self).__init__()
        self.scaling_fn = scaling_fn
        self.decoder = BaseLSTMDecoder(params, output_dim + known_future_covariates)
        self.distribution = Distribution(params.hidden_dim, output_dim)
        self.known_future_covariates = known_future_covariates
        self.output_dim = output_dim

    def forward(
        self, X: torch.tensor, pad_mask: torch.tensor, conts_future: torch.tensor = None
    ) -> torch.tensor:
        X = X.clone()
        B, T, _ = X.shape
        target_scale = self.scaling_fn(X, pad_mask)
        conts = X / target_scale
        conts_list = [conts, torch.log(target_scale[:, :, :1]).expand(B, T, 1)]

        if conts_future is not None:
            conts_list.append(conts_future)

        embedded = torch.cat(conts_list, 2)
        out, _ = self.decoder(embedded)
        distr = self.distribution(out, target_scale=target_scale)
        return distr

    def teacher_forcing_output(
        self, batch: datasets.PaddedData
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data
        conts_future = x[:, :, -self.known_future_covariates :]
        x = x[:, :, : -self.known_future_covariates]

        pad_mask = batch.pad_mask
        y_hat = self(
            x[:, :-1, :], pad_mask[:, :-1], conts_future=conts_future[:, :-1, :]
        )
        y = x[:, 1:, :1]

        y_hat = y_hat[pad_mask[:, 1:]]
        y = y[pad_mask[:, 1:]]
        return y_hat, y

    def generate(
        self,
        X: torch.tensor,
        pad_mask: torch.tensor,
        prediction_length: int,
        context_length: int,
    ) -> torch.tensor:
        y_hat = torch.zeros(X.shape[0], prediction_length, self.output_dim)
        past_conts = X[:, :context_length, :]
        past_pad_mask = pad_mask[:, :context_length]

        future_conts = X[:, context_length:, -self.known_future_covariates :]
        future_pad_mask = pad_mask[:, context_length:]

        for x in range(prediction_length):
            preds = self(
                past_conts[:, :, : -self.known_future_covariates],
                past_pad_mask,
                conts_future=past_conts[:, :, -self.known_future_covariates :],
            )
            preds = preds[:, -1:, :].detach().clone()  # becomes (B, 1, C)
            y_hat[:, x, :] = preds[:, 0, :]

            # append the prediction to the past
            past_conts[:, :, : -self.known_future_covariates] = torch.cat(
                (past_conts[:, :, : -self.known_future_covariates], preds), dim=1
            )[:, 1:, :]
            past_conts[:, :, -self.known_future_covariates :] = torch.cat(
                [
                    past_conts[:, :, -self.known_future_covariates :],
                    future_conts[:, x : x + 1, :],
                ],
                dim=1,
            )[:, 1:, :]
            past_pad_mask = torch.cat(
                [past_pad_mask, future_pad_mask[:, x : x + 1]], dim=1
            )[:, 1:]

        return y_hat

    def validate(self, batch: datasets.PaddedData, prediction_length, context_length):
        y = batch.data[
            :, context_length : context_length + prediction_length, : self.output_dim
        ]
        y_hat = self.generate(
            batch.data,
            batch.pad_mask,
            prediction_length=prediction_length,
            context_length=context_length,
        )
        y_hat = y_hat[:, -prediction_length:, ...]
        return y_hat, y



# Third Party
from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor, nn
from transformers import (
    EarlyStoppingCallback,
    PatchTSMixerConfig,
    PatchTSMixerForPrediction,
    PatchTSMixerModel,
    Trainer,
    TrainingArguments,
)

from nnts.utils import Scheduler, TrainingMethod

from ..datasets import PaddedData


class PatchTSMixerForPredictionHead(nn.Module):
    """Prediction Head for Forecasting

    Args:
        config (`PatchTSMixerConfig`):
            Configuration.
    """

    def __init__(self, config: PatchTSMixerConfig, distribution_output=None):
        super().__init__()

        self.prediction_channel_indices = config.prediction_channel_indices

        if self.prediction_channel_indices is not None:
            self.prediction_channel_indices.sort()

        self.dropout_layer = nn.Dropout(config.head_dropout)
        if distribution_output is None:
            self.base_forecast_block = nn.Linear(
                (config.num_patches * config.d_model), config.prediction_length
            )
        else:
            self.base_forecast_block = distribution_output.get_parameter_projection(
                config.num_patches * config.d_model
            )

        self.flatten = nn.Flatten(start_dim=-2)

    def forward(self, hidden_features):
        """

        Args:
            hidden_features (`torch.Tensor` of shape `(batch_size, num_patch, d_model)` in `flatten` mode
                or `(batch_size, n_vars, num_patch, d_model)` in `common_channel`/`mix_channel` mode.): Input hidden
                features.

        Returns:
            `torch.Tensor` of shape `(batch_size, prediction_length, nvars)`.

        """

        hidden_features = self.flatten(
            hidden_features
        )  # [batch_size x n_vars x num_patch * d_model]
        hidden_features = self.dropout_layer(
            hidden_features
        )  # [batch_size x n_vars x num_patch * d_model]
        forecast = self.base_forecast_block(
            hidden_features
        )  # [batch_size x n_vars x prediction_length]
        if isinstance(forecast, tuple):
            forecast = tuple(z.transpose(-1, -2) for z in forecast)
        else:
            forecast = forecast.transpose(
                -1, -2
            )  # [batch_size x prediction_length x n_vars]

        if self.prediction_channel_indices is not None:
            if isinstance(forecast, tuple):
                forecast = tuple(
                    z[..., self.prediction_channel_indices] for z in forecast
                )
            else:
                forecast = forecast[
                    ..., self.prediction_channel_indices
                ]  # [batch_size x prediction_length x n_vars]

        return forecast


@dataclass
class Hyperparams:
    patch_length: int
    num_input_channels: int
    optimizer: callable = torch.optim.Adam
    loss_fn: callable = F.l1_loss
    dropout: float = 0.0
    batch_size: int = 32
    lr: float = 0.005
    epochs: int = 100
    patience: int = 10
    early_stopper_patience: int = 30
    batches_per_epoch: int = 50
    weight_decay: float = 0.0
    training_method: TrainingMethod = TrainingMethod.DMS
    scheduler: Scheduler = Scheduler.REDUCE_LR_ON_PLATEAU
    model_file_path: str = f"logs"
    d_model: int = 48
    num_layers: int = 3
    expansion_factor: int = 3
    dropout: float = 0.5
    head_dropout: float = 0.7
    mode: str = (
        "common_channel"  # change it `mix_channel` if we need to explicitly model channel correlations
    )
    scaling: str = "std"


class PatchTSMixer(PatchTSMixerModel):

    def __init__(
        self, context_length: int, prediction_length: int, params: Hyperparams
    ):

        config = PatchTSMixerConfig(
            context_length=context_length,
            prediction_length=prediction_length,
            patch_length=params.patch_length,
            num_input_channels=params.num_input_channels,
            patch_stride=params.patch_length,
            d_model=params.d_model,
            num_layers=params.num_layers,
            expansion_factor=params.expansion_factor,
            dropout=params.dropout,
            head_dropout=params.head_dropout,
            mode=params.mode,  # change it `mix_channel` if we need to explicitly model channel correlations
            scaling=params.scaling,
        )
        super().__init__(config)
        self.head = PatchTSMixerForPredictionHead(
            config=config,
            distribution_output=None,
        )
        self.seq_len = context_length

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data[:, : self.seq_len, :]
        pad_mask = batch.pad_mask[:, : self.seq_len]
        model_output = self(x)
        y_hat = self.head(model_output.last_hidden_state)
        y_hat = y_hat * model_output.scale + model_output.loc

        y = batch.data[:, self.seq_len :, :]

        pad_mask = batch.pad_mask[:, self.seq_len :]
        y_hat = y_hat[pad_mask]
        y = y[pad_mask]

        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        model_output = self(x)
        y_hat = self.head(model_output.last_hidden_state)
        y_hat = y_hat * model_output.scale + model_output.loc
        return y_hat



from dataclasses import dataclass
from typing import Callable, Iterable, List, Tuple, Type

import pandas as pd
import torch
from torch.utils.data import DataLoader

import nnts.preprocessing
import nnts.utils


def count_of_params_in(net: torch.nn.Module):
    pytorch_total_params = sum(p.numel() for p in net.parameters() if p.requires_grad)
    return pytorch_total_params


class DataLoaderBuilder:

    def __init__(self, data):
        self.data = data
        self.transforms = None
        self.Dataset = None
        self.dataset_options = {}
        self.DataLoader = None
        self.dataloader_options = {}
        self.Sampler = None

    def set_transforms(
        self, transforms: List[nnts.preprocessing.Transformation]
    ) -> "DataLoaderBuilder":
        self.transforms = transforms
        return self

    def set_dataset(self, Dataset: Type, dataset_options) -> "DataLoaderBuilder":
        self.Dataset = Dataset
        self.dataset_options = dataset_options
        return self

    def set_dataloader(
        self, DataLoader: Type, **dataloader_options
    ) -> "DataLoaderBuilder":
        self.DataLoader = DataLoader
        self.dataloader_options = dataloader_options
        return self

    def set_Sampler(self, Sampler: Type) -> "DataLoaderBuilder":
        self.Sampler = Sampler
        return self

    def to_dataloader(self) -> Iterable:
        if self.transforms is not None:
            for transform in self.transforms:
                self.data = transform.transform(self.data)

        ts = self.Dataset(self.data, **self.dataset_options).build()

        if self.Sampler is not None:
            sampler = self.Sampler(ts)
            self.dataloader_options["sampler"] = sampler
            self.dataloader_options.pop("shuffle", None)

        dl = self.DataLoader(ts, **self.dataloader_options)
        return dl


def create_dataloaders(
    df: pd.DataFrame,
    splitter_fn: Callable,
    context_length: int,
    prediction_length: int,
    Dataset: Type,
    dataset_options: dict = {},
    Sampler: Type | None = None,
    batch_size: int = 32,
    transforms: List[nnts.preprocessing.Transformation] = None,
) -> Tuple[Iterable, Iterable, Iterable]:
    """takes split data -> fits transforms -> creates dataloaders"""
    split_data = splitter_fn(df, context_length, prediction_length)
    return create_dataloaders_from_split_data(
        split_data,
        Dataset,
        dataset_options,
        Sampler,
        batch_size,
        transforms,
    )


def create_dataloaders_from_split_data(
    split_data: Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame],
    Dataset: Type,
    dataset_options: dict = {},
    Sampler: Type | None = None,
    batch_size: int = 32,
    transforms: List[nnts.preprocessing.Transformation] = None,
) -> Tuple[Iterable, Iterable, Iterable]:
    """takes split data -> fits transforms -> creates dataloaders"""

    has_validation = hasattr(split_data, "validation")

    trn_builder = DataLoaderBuilder(split_data.train)
    test_builder = DataLoaderBuilder(split_data.test)
    if has_validation:
        val_builder = DataLoaderBuilder(split_data.validation)

    if transforms is not None:
        transforms = [
            transform.fit(
                split_data.train,
            )
            for transform in transforms
        ]
        trn_builder.set_transforms(transforms)
        test_builder.set_transforms(transforms)
        if has_validation:
            val_builder.set_transforms(transforms)

    if Sampler is not None:
        trn_builder.set_Sampler(Sampler)
        # test_builder.set_Sampler(Sampler)
        if has_validation:
            val_builder.set_Sampler(Sampler)

    trn_dl = (
        trn_builder.set_dataset(
            Dataset,
            dataset_options=dataset_options,
        )
        .set_dataloader(DataLoader, batch_size=batch_size, shuffle=True, drop_last=True)
        .to_dataloader()
    )
    test_dl = (
        test_builder.set_dataset(
            Dataset,
            dataset_options=dataset_options,
        )
        .set_dataloader(DataLoader, batch_size=batch_size, shuffle=False)
        .to_dataloader()
    )
    if has_validation:
        val_dl = (
            val_builder.set_dataset(
                Dataset,
                dataset_options=dataset_options,
            )
            .set_dataloader(
                DataLoader, batch_size=batch_size, shuffle=False, drop_last=False
            )
            .to_dataloader()
        )
        return trn_dl, val_dl, test_dl
    return trn_dl, test_dl


def seed_everything(seed: int):
    import os
    import random

    import numpy as np
    import torch

    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True



from .autoformer import Autoformer
from .baselstm import BaseFutureCovariateLSTM, BaseLSTM, LinearModel
from .deepar import DeepARPoint, DistrDeepAR
from .dlinear import DLinear
from .itransformer import iTransformer
from .nbeats import NBeats
from .nhits import NHITS
from .nlinear import NLinear
from .patchtst import PatchTST
from .seglstm import SegLSTM
from .tide import TiDE
from .timemixer import TimeMixer
from .timesnet import TimesNet
from .tsmixer import TSMixer
from .unrolledlstm import UnrolledLSTM

__all__ = [
    "LinearModel",
    "BaseLSTM",
    "SegLSTM",
    "UnrolledLSTM",
    "BaseFutureCovariateLSTM",
    "DeepARPoint",
    "DistrDeepAR",
    "NLinear",
    "DLinear",
    "NHITS",
    "TiDE",
    "PatchTST",
    "Autoformer",
    "NBeats",
    "TSMixer",
    "TimeMixer",
    "iTransformer",
    "TimesNet",
]



from typing import List

import pandas as pd
import torch

import nnts.data
import nnts.preprocessing


def masked_mean_abs_scaling(
    seq: torch.Tensor,
    mask: torch.Tensor | None = None,
    eps: float = 1e-10,
    dim: int = 1,
):
    if mask is None:
        mask = torch.ones_like(seq)
    if len(mask.shape) == 2 and len(seq.shape) == 3:
        mask = mask[:, :, None]

    if len(mask.shape) != len(seq.shape):
        raise ValueError(
            f"Mask shape {mask.shape} does not match sequence shape {seq.shape}"
        )
    seq_sum = (seq * mask).abs().sum(dim, keepdim=True)
    item_count = mask.sum(dim, keepdim=True).clamp(min=1)

    scale = seq_sum / item_count

    if scale.max() < 1:
        scale = torch.clamp(scale, min=eps)
    else:
        scale = torch.clamp(scale, min=1)
    return scale


class StandardScaler(nnts.preprocessing.Transformation):

    def __init__(self, mean=None, std=None, cols=None):
        self.mean = mean
        self.std = std
        self.cols = cols

    def fit(self, data: pd.DataFrame, cols=None):
        if self.cols is not None:
            cols = self.cols

        numeric_data = (
            data.select_dtypes(include=["number"]) if cols is None else data[cols]
        )
        self.mean = numeric_data.mean()
        self.std = numeric_data.std()
        return self

    def transform(self, data: pd.DataFrame, cols=None):
        if self.cols is not None:
            cols = self.cols

        numeric_data = (
            data.select_dtypes(include=["number"]) if cols is None else data[cols]
        )
        numeric_cols = numeric_data.columns
        data[numeric_cols] = (numeric_data - self.mean) / self.std
        return data

    def inverse_transform(self, data: pd.DataFrame, cols=None):
        if self.cols is not None:
            cols = self.cols
        numeric_data = (
            data.select_dtypes(include=["number"]) if cols is None else data[cols]
        )
        numeric_cols = numeric_data.columns
        data[numeric_cols] = numeric_data * self.std + self.mean
        return data


class MaxMinScaler(nnts.preprocessing.Transformation):
    def __init__(self, max=None, min=None, cols=None):
        self.max = max
        self.min = min
        self.cols = cols

    def fit(self, data: pd.DataFrame, cols=None):
        if self.cols is not None:
            cols = self.cols
        numeric_data = (
            data.select_dtypes(include=["number"]) if cols is None else data[cols]
        )
        self.max = numeric_data.max()
        self.min = numeric_data.min()
        return self

    def transform(self, data: pd.DataFrame, cols=None):
        if self.cols is not None:
            cols = self.cols
        numeric_data = (
            data.select_dtypes(include=["number"]) if cols is None else data[cols]
        )
        numeric_cols = numeric_data.columns
        data[numeric_cols] = ((numeric_data - self.min) / (self.max - self.min)) - 0.5
        return data

    def inverse_transform(self, data: pd.DataFrame, cols=None):
        if self.cols is not None:
            cols = self.cols
        numeric_data = (
            data.select_dtypes(include=["number"]) if cols is None else data[cols]
        )
        numeric_cols = numeric_data.columns
        data[numeric_cols] = numeric_data * (self.max - self.min) + self.min
        return data


def create_lags(
    n_timesteps: int, past_target: torch.tensor, lag_seq: List[int]
) -> torch.tensor:
    lag_features = []
    for t in range(0, n_timesteps):
        lag_seq = lag_seq + 1
        lag_for_step = past_target.index_select(1, lag_seq)
        lag_features.append(lag_for_step)
    return torch.stack(lag_features, dim=1).flip(1)


class LagProcessor:

    def __init__(self, lag_seq: List[int]):
        self.lag_seq = torch.tensor(lag_seq)

    def create(self, n_timesteps: int, past_target: torch.tensor):
        return create_lags(n_timesteps, past_target, self.lag_seq - 1)

    def __len__(self):
        return len(self.lag_seq)

    def max(self):
        return max(self.lag_seq)



from typing import Dict, List

import torch
import torch.distributions as td
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import AffineTransform, Distribution, TransformedDistribution

import nnts.data
import nnts.torch.preprocessing
from nnts import utils

from .. import datasets, models

FEAT_SCALE: str = "feat_scale"


def distr_nll(distr: td.Distribution, target: torch.Tensor) -> torch.Tensor:
    nll = -distr.log_prob(target)
    return nll.mean()


class AffineTransformed(TransformedDistribution):
    """
    Represents the distribution of an affinely transformed random variable.

    This is the distribution of ``Y = scale * X + loc``, where ``X`` is a
    random variable distributed according to ``base_distribution``.

    Parameters
    ----------
    base_distribution
        Original distribution
    loc
        Translation parameter of the affine transformation.
    scale
        Scaling parameter of the affine transformation.
    """

    def __init__(self, base_distribution: Distribution, loc=None, scale=None):
        self.scale = 1.0 if scale is None else scale
        self.loc = 0.0 if loc is None else loc

        super().__init__(base_distribution, [AffineTransform(self.loc, self.scale)])

    @property
    def mean(self):
        """
        Returns the mean of the distribution.
        """
        return self.base_dist.mean * self.scale + self.loc

    @property
    def variance(self):
        """
        Returns the variance of the distribution.
        """
        return self.base_dist.variance * self.scale**2

    @property
    def stddev(self):
        """
        Returns the standard deviation of the distribution.
        """
        return self.variance.sqrt()


class StudentTHead(nn.Module):
    """
    This model outputs a studentT distribution.
    """

    PARAMS = 3

    def __init__(self, hidden_size: int, output_size: int):
        super(StudentTHead, self).__init__()

        self.main = nn.ModuleList(
            [nn.Linear(hidden_size, output_size) for _ in range(StudentTHead.PARAMS)]
        )

    def forward(self, x: torch.tensor, target_scale: torch.tensor):
        df, loc, scale = tuple(self.main[i](x) for i in range(StudentTHead.PARAMS))
        df = 2.0 + F.softplus(df)
        epsilon = torch.finfo(scale.dtype).eps
        scale = F.softplus(scale).clamp_min(epsilon)
        return df, loc, scale


class DeepARMixin:
    def generate(
        self,
        X: torch.tensor,
        pad_mask: torch.tensor,
        prediction_length: int,
        context_length: int,
    ) -> torch.tensor:
        y_hat = self(X, pad_mask, prediction_length, context_length)
        y_hat = y_hat[:, -prediction_length:, :]
        return y_hat

    def validate(self, batch: datasets.PaddedData, prediction_length, context_length):
        y = batch.data[:, -prediction_length:, : self.output_dim]
        y_hat = self.generate(
            batch.data,
            batch.pad_mask,
            prediction_length=prediction_length,
            context_length=context_length,
        )
        return y_hat, y

    def free_running(
        self, batch: datasets.PaddedData, prediction_length: int, context_length: int
    ) -> torch.tensor:
        x = batch.data
        pad_mask = batch.pad_mask
        y = x[:, 1 - context_length - prediction_length :, : self.output_dim]
        y_hat = self(
            x,
            pad_mask,
            prediction_length,
            context_length,
        )
        return y_hat, y

    def prep_input(self, X, context_length, past_target, target_scale):
        B, T, _ = X.shape
        features = []
        if self.lag_processor is not None:
            lags = torch.zeros(B, T, len(self.lag_processor))
            lags[:, :context_length, :] = self.lag_processor.create(
                context_length, past_target
            )
            features.append(lags)

        if FEAT_SCALE in self.scaled_features:
            features.append(
                torch.log(target_scale[:, :, :1]).expand(B, T, 1),
            )

        if self.cat_idx is not None:
            X[..., self.cat_idx] = self.embedder(X[..., self.cat_idx].long()).squeeze(
                -1
            )

        if self.seq_cat_idx is not None:
            features.append(self.seq_embedder(X[..., self.seq_cat_idx].long()))
            X = torch.cat(
                [X[..., : self.seq_cat_idx], X[..., self.seq_cat_idx + 1 :]], 2
            )

        input = torch.cat([X[:, :, :1]] + features + [X[:, :, 1:]], 2)
        return lags, input


class DeepARPoint(nn.Module, DeepARMixin):

    def __init__(
        self,
        Distribution: nn.Module,
        params: utils.Hyperparams,
        scaling_fn: callable,
        output_dim: int,
        lag_processor: nnts.torch.preprocessing.LagProcessor,
        scaled_features: List[str],
        context_length: int = 15,
        cat_idx: int = None,
        seq_cat_idx=None,
        emb_dim=5,
    ):
        super(DeepARPoint, self).__init__()
        self.scaling_fn = scaling_fn
        self.output_dim = output_dim
        self.scaled_features = scaled_features
        self.n_scaled_features = len(scaled_features)
        self.n_scaled_features_excluding_feat_scale = len(
            [f for f in scaled_features if f != FEAT_SCALE]
        )
        self.lag_processor = lag_processor
        self.decoder = models.unrolledlstm.UnrolledLSTMDecoder(
            params, self.n_scaled_features + len(self.lag_processor) + emb_dim - 1
        )
        self.distribution = Distribution(params.hidden_dim, output_dim)
        self.context_length = context_length
        self.cat_idx = cat_idx
        if cat_idx is not None:
            self.embedder = nn.Embedding(1, 1)

        self.seq_cat_idx = seq_cat_idx
        if seq_cat_idx is not None:
            self.seq_embedder = nn.Embedding(12, emb_dim)

    def forward(
        self,
        X: torch.tensor,
        pad_mask: torch.tensor,
        H: int,
        context_length: int,
    ) -> torch.tensor:
        X = X.clone()
        if H > 0:
            X[:, -H:, 0] = 0.0
            past_target = X[:, :-H, 0].flip(1)
        else:
            past_target = X[:, :, 0].flip(1)
        X = X[:, self.lag_processor.max() :, ...]
        pad_mask = pad_mask[:, self.lag_processor.max() :]
        B, T, C = X.shape

        target_scale = self.scaling_fn(
            X[:, : self.context_length, :1], pad_mask[:, : self.context_length]
        )
        X[:, :, :1] = X[:, :, :1] / target_scale
        past_target = past_target / target_scale.squeeze(2)
        _, input = self.prep_input(X, context_length, past_target, target_scale)

        if H > 0:
            input = input[:, :-1, :]
            y_hat = torch.zeros(B, T - 1, self.output_dim)
        else:
            y_hat = torch.zeros(B, T, self.output_dim)

        out, hidden = self.decoder(input[:, :context_length, :])
        out = self.distribution(out, target_scale=None)

        y_hat[:, :context_length, :] = out

        # free running for H steps
        for t in range(0, H - 1):
            past_target = torch.cat([out[:, -1:, 0], past_target], 1)
            lags = self.lag_processor.create(1, past_target)
            input[..., context_length + t, 0] = out[:, -1, 0]
            input[..., context_length + t, 1 : 1 + lags.shape[2]] = lags.squeeze(1)

            # select the lag features and detach from the graph to prevent backpropagation
            out, hidden = self.decoder(
                input[:, context_length + t : context_length + t + 1, ...]
                .clone()
                .detach(),
                hidden,
            )
            out = self.distribution(out, target_scale=None)
            y_hat[:, context_length + t, :] = out[:, -1, :]

        y_hat = y_hat * target_scale[:, :, : self.output_dim]
        return y_hat

    def teacher_forcing_output(
        self, batch: datasets.PaddedData, prediction_length, context_length
    ):
        """
        data: dict with keys "X" and "pad_mask"
        """
        x = batch.data
        pad_mask = batch.pad_mask
        y = x[:, 1:, : self.output_dim]
        y_hat = self(
            x[:, :-1, :], pad_mask[:, :-1], 0, context_length + prediction_length - 1
        )
        y = y[:, 1 - context_length - prediction_length :, ...]
        return y_hat, y


class DistrDeepAR(nn.Module, DeepARMixin):

    def __init__(
        self,
        Distribution: nn.Module,
        params: utils.Hyperparams,
        scaling_fn: callable,
        output_dim: int,
        lag_processor: nnts.torch.preprocessing.LagProcessor,
        scaled_features: List[str],
        context_length: int = 15,
        cat_idx: int = None,
        seq_cat_idx=None,
        emb_dim=1,
    ):
        super(DistrDeepAR, self).__init__()

        self.scaling_fn = scaling_fn
        self.output_dim = output_dim
        self.scaled_features = scaled_features
        self.n_scaled_features = len(scaled_features)
        self.n_scaled_features_excluding_feat_scale = len(
            [f for f in scaled_features if f != FEAT_SCALE]
        )
        self.lag_processor = lag_processor
        self.decoder = models.unrolledlstm.UnrolledLSTMDecoder(
            params, self.n_scaled_features + len(self.lag_processor) + emb_dim - 1
        )
        self.distribution = Distribution(params.hidden_dim, output_dim)

        self.context_length = context_length
        self.cat_idx = cat_idx
        if cat_idx is not None:
            self.embedder = nn.Embedding(1, 1)

        self.seq_cat_idx = seq_cat_idx
        if seq_cat_idx is not None:
            self.seq_embedder = nn.Embedding(12, emb_dim)

    def _distr(
        self,
        X: torch.tensor,
        pad_mask: torch.tensor,
        H: int,
        context_length: int,
    ) -> torch.tensor:
        X = X.clone()
        if H > 0:
            X[:, -H:, 0] = 0.0
            past_target = X[:, :-H, 0].flip(1)
        else:
            past_target = X[:, :, 0].flip(1)
        X = X[:, self.lag_processor.max() :, ...]
        pad_mask = pad_mask[:, self.lag_processor.max() :]

        target_scale = self.scaling_fn(
            X[:, : self.context_length, :1], pad_mask[:, : self.context_length]
        )
        X[:, :, :1] = X[:, :, :1] / target_scale
        past_target = past_target / target_scale.squeeze(2)
        _, input = self.prep_input(X, context_length, past_target, target_scale)

        if H > 0:
            input = input[:, :-1, :]

        out, _ = self.decoder(input[:, :context_length, :])
        params = self.distribution(out, target_scale=target_scale)
        distr = AffineTransformed(
            td.StudentT(params[0], params[1], params[2]), None, target_scale
        )

        return distr

    def forward(
        self,
        X: torch.tensor,
        pad_mask: torch.tensor,
        H: int,
        context_length: int,
        n_samples: int = 100,
    ) -> torch.tensor:
        X = X.clone()
        if H > 0:
            X[:, -H:, 0] = 0.0
            past_target = X[:, :-H, 0].flip(1)
        else:
            past_target = X[:, :, 0].flip(1)
        X = X[:, self.lag_processor.max() :, ...]
        pad_mask = pad_mask[:, self.lag_processor.max() :]
        B, T, C = X.shape

        target_scale = self.scaling_fn(
            X[:, :context_length, :1], pad_mask[:, :context_length]
        )
        X[:, :, :1] = X[:, :, :1] / target_scale
        past_target = past_target / target_scale.squeeze(2)

        lags, input = self.prep_input(X, context_length, past_target, target_scale)
        input = input[:, :-1, :]

        out, hidden = self.decoder(input[:, :context_length, :])

        # Expand the input and target_scale to N samples
        params = self.distribution(out, target_scale=target_scale)

        params = tuple(
            param.repeat_interleave(n_samples, 0) for param in params
        )  # N*B, T, C

        input = input.repeat_interleave(n_samples, 0)  # N*B, T, C
        past_target = past_target.repeat_interleave(n_samples, 0)  # N*B, T
        target_scale = target_scale.repeat_interleave(n_samples, 0)  # N*B, T, 1
        lags = lags.repeat_interleave(n_samples, 0)  # N*B, T, L
        hidden = tuple(
            state.repeat_interleave(n_samples, 1) for state in hidden
        )  # N*B, H
        distr = AffineTransformed(
            td.StudentT(params[0], params[1], params[2]), None, target_scale
        )
        out = distr.sample()  # N*B, T

        y_hat = torch.zeros(n_samples * B, T - 1, 1)
        y_hat[..., :context_length, :] = out

        # free running for H steps
        for t in range(0, H - 1):
            out[:, -1:, 0] = out[:, -1:, 0] / target_scale.squeeze(2)
            past_target = torch.cat([out[:, -1:, 0], past_target], 1)
            lags = self.lag_processor.create(1, past_target)
            input[..., context_length + t, 0] = out[:, -1, 0]
            input[..., context_length + t, 1 : 1 + lags.shape[2]] = lags.squeeze(1)

            # select the lag features and detach from the graph to prevent backpropagation
            out, hidden = self.decoder(
                input[:, context_length + t : context_length + t + 1, ...]
                .clone()
                .detach(),
                hidden,
            )
            params = self.distribution(out, target_scale=target_scale)
            distr = AffineTransformed(
                td.StudentT(params[0], params[1], params[2]),
                None,
                target_scale,
            )
            out = distr.sample()
            y_hat[..., context_length + t, :] = out[:, -1, :]

        y_hat = y_hat.reshape(B, n_samples, -1).permute(1, 0, 2)
        y_hat = y_hat.median(dim=0)[0]
        return y_hat.unsqueeze(-1)

    def teacher_forcing_output(
        self, batch: datasets.PaddedData, prediction_length, context_length
    ):
        """
        data: dict with keys "X" and "pad_mask"
        """
        x = batch.data
        pad_mask = batch.pad_mask
        y = x[:, 1:, : self.output_dim]
        y_hat = self._distr(
            x[:, :-1, :], pad_mask[:, :-1], 0, context_length + prediction_length - 1
        )
        y = y[:, 1 - context_length - prediction_length :, ...]
        return y_hat, y



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any


class Transformation(ABC):
    @abstractmethod
    def transform(self, data: Any, *args, **kwargs) -> Any:
        pass

    @abstractmethod
    def fit(self, data, *args, **kwargs) -> Transformation:
        pass

    @abstractmethod
    def inverse_transform(self, data, *args, **kwargs) -> Any:
        pass



from collections import namedtuple
from typing import Iterator, List, Optional, Sized

import pandas as pd
import torch
import torch.utils
import torch.utils.data
from torch.utils.data import Sampler

PaddedData = namedtuple("PaddedData", ["data", "pad_mask"])


def right_pad_sequence(
    seq: list[torch.Tensor], padding_value: int = 0, min_length: int = 0
):
    """Pads a list of 2D tensors to the right with a given padding value."""
    max_lengths = max([t.shape[0] for t in seq])
    max_lengths = max(max_lengths, min_length)
    padded_tensor = torch.zeros(len(seq), max_lengths, seq[0].shape[1]) + padding_value
    padded_mask = torch.zeros(len(seq), max_lengths).bool()
    # padded_mask[:, :min_length] = True
    for i in range(len(seq)):
        start = min_length - seq[i].shape[0]
        start = max(0, start)
        padded_tensor[i, start : start + seq[i].shape[0], ...] = seq[i]
        padded_mask[i, start : start + seq[i].shape[0]] = True

    return padded_tensor, padded_mask


def left_pad_sequence(
    seq: list[torch.Tensor], padding_value: int = 0, min_length: int = 0
):
    """Pads a list of 2D tensors to the left with a given padding value."""
    max_lengths = max([t.shape[0] for t in seq])
    max_lengths = max(max_lengths, min_length)
    padded_tensor = torch.full((len(seq), max_lengths, seq[0].shape[1]), padding_value)
    padded_mask = torch.zeros(len(seq), max_lengths).bool()
    for i in range(len(seq)):
        pad_len = max_lengths - seq[i].shape[0]
        pad_len = max(pad_len, min_length - seq[i].shape[0])
        padded_tensor[i, pad_len : pad_len + seq[i].shape[0], ...] = seq[i]
        padded_mask[i, pad_len : pad_len + seq[i].shape[0]] = True

    return padded_tensor.float(), padded_mask


class TimeseriesDataset(torch.utils.data.Dataset):
    """Time series dataset for use with Global models where given a
    a dataset containing multiple time series we want to sample windows
    of a fixed length determined by the context_length and prediction_length
    and optionally lag length for each time series.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        context_length: int,
        prediction_length: int,
        conts: List[int] = [],
        lag_seq: List[int] = None,
        pad_fn=right_pad_sequence,
    ):
        self.df = df.copy()
        self.conts = conts.copy()

        self.lag_length = 0 if lag_seq is None else max(lag_seq)
        self.context_length = torch.tensor(context_length).long()
        self.prediction_length = torch.tensor(prediction_length).long()

        lengths = (
            self.df.groupby("unique_id", sort=False)["ds"].count()
            - context_length
            - prediction_length
            - self.lag_length
            + 1  # we return x and y so add 1 for teacher forcing
        ).clip(1)
        cum_lengths = lengths.cumsum()
        self.shifted_cum_lengths = torch.tensor(
            cum_lengths.shift().fillna(0).astype(int).values
        )
        self.len = lengths.sum()
        self.min_length = context_length + prediction_length + self.lag_length
        self.lengths = lengths
        self.pad_fn = pad_fn

    def build(self) -> "TimeseriesDataset":
        ts = []
        for unique_id, grp in self.df.groupby("unique_id", sort=False):
            ts.append(torch.from_numpy(grp[["y"] + self.conts].values))
        ts, mask = self.pad_fn(ts, min_length=self.min_length)
        self.X = ts[:, :, :]
        self.pad_mask = mask
        return self

    def __len__(self) -> int:
        return self.len

    def __getitem__(self, i: int) -> PaddedData:
        shifted_ln_lt_i = self.shifted_cum_lengths[self.shifted_cum_lengths <= i]
        result = torch.max(shifted_ln_lt_i, 0)
        locator = result.indices, i - result.values
        X = self.X[
            locator[0],
            locator[1] : locator[1]
            + self.lag_length
            + self.context_length
            + self.prediction_length,
            ...,
        ]
        pad_mask = self.pad_mask[
            locator[0],
            locator[1] : locator[1]
            + self.lag_length
            + self.context_length
            + self.prediction_length,
        ]
        return PaddedData(data=X, pad_mask=pad_mask)


class TimeSeriesSampler(Sampler[int]):

    data_source: Sized

    def __init__(self, data_source: Sized, num_samples: Optional[int] = None):
        self.data_source = data_source
        self._num_samples = len(data_source) if num_samples is None else num_samples

        to = torch.cat(
            [
                data_source.shifted_cum_lengths,
                torch.tensor([len(data_source)]),
            ]
        )[1:]
        self.ranges = torch.stack([data_source.shifted_cum_lengths, to]).T

    def __iter__(self) -> Iterator[int]:
        count = 0
        while True:
            for i in range(self.ranges.shape[0]):
                count += 1
                if count > self._num_samples:
                    return
                start, end = self.ranges[i]
                val = torch.randint(start, end, size=(1,))
                yield val

    def __len__(self) -> int:
        return self._num_samples


class TimeseriesDataset(torch.utils.data.Dataset):
    """Time series dataset for use with Global models where given a
    a dataset containing multiple time series we want to sample windows
    of a fixed length determined by the context_length and prediction_length
    and optionally lag length for each time series.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        context_length: int,
        prediction_length: int,
        conts: List[int] = [],
        lag_seq: List[int] = None,
        pad_fn=right_pad_sequence,
    ):
        self.df = df.copy()
        self.conts = conts.copy()

        self.lag_length = 0 if lag_seq is None else max(lag_seq)
        self.context_length = torch.tensor(context_length).long()
        self.prediction_length = torch.tensor(prediction_length).long()

        lengths = (
            self.df.groupby("unique_id", sort=False)["ds"].count()
            - context_length
            - prediction_length
            - self.lag_length
            + 1  # we return x and y so add 1 for teacher forcing
        ).clip(1)
        cum_lengths = lengths.cumsum()
        self.shifted_cum_lengths = torch.tensor(
            cum_lengths.shift().fillna(0).astype(int).values
        )
        self.len = lengths.sum()
        self.min_length = context_length + prediction_length + self.lag_length
        self.lengths = lengths
        self.pad_fn = pad_fn

    def build(self) -> "TimeseriesDataset":
        ts = []
        for unique_id, grp in self.df.groupby("unique_id", sort=False):
            ts.append(torch.from_numpy(grp[["y"] + self.conts].values))
        ts, mask = self.pad_fn(ts, min_length=self.min_length)
        self.X = ts[:, :, :]
        self.pad_mask = mask
        return self

    def __len__(self) -> int:
        return self.len

    def __getitem__(self, i: int) -> PaddedData:
        shifted_ln_lt_i = self.shifted_cum_lengths[self.shifted_cum_lengths <= i]
        result = torch.max(shifted_ln_lt_i, 0)
        locator = result.indices, i - result.values
        X = self.X[
            locator[0],
            locator[1] : locator[1]
            + self.lag_length
            + self.context_length
            + self.prediction_length,
            ...,
        ]
        pad_mask = self.pad_mask[
            locator[0],
            locator[1] : locator[1]
            + self.lag_length
            + self.context_length
            + self.prediction_length,
        ]
        return PaddedData(data=X, pad_mask=pad_mask)


class MultivariateTimeSeriesDatasetLong(torch.utils.data.Dataset):
    """Time series dataset for use with Multivariate or Local models where given a
    a dataset containing multiple time series, N, we want to sample windows
    of a fixed length determined by the context_length and prediction_length
    across all time series so each sample will have shape (N, context_length + prediction_length)
    """

    def __init__(
        self,
        df: pd.DataFrame,
        context_length: int,
        prediction_length: int,
        conts: List[int] = [],
        lag_seq: List[int] = None,
        pad_fn=right_pad_sequence,
    ):
        self.df = df.copy()
        self.conts = conts.copy()

        self.lag_length = 0 if lag_seq is None else max(lag_seq)
        self.context_length = torch.tensor(context_length).long()
        self.prediction_length = torch.tensor(prediction_length).long()

        lengths = (
            self.df.groupby("unique_id", sort=False)["ds"].count()
            - context_length
            - prediction_length
            - self.lag_length
            + 1  # we return x and y so add 1 for teacher forcing
        ).clip(1)
        self.max_length = lengths.max()
        self.window_size = context_length + prediction_length + self.lag_length
        self.lengths = lengths
        self.pad_fn = pad_fn

    def build(self) -> "MultivariateTimeSeriesDataset":
        ts = []
        for unique_id, grp in self.df.groupby("unique_id", sort=False):
            ts.append(torch.from_numpy(grp[["y"] + self.conts].values))
        ts, mask = self.pad_fn(ts, min_length=self.window_size)
        self.X = ts[:, :, :]
        self.pad_mask = mask
        return self

    def __len__(self) -> int:
        return self.max_length

    def __getitem__(self, i: int) -> PaddedData:
        X = self.X[:, i : i + self.window_size, 0].permute(1, 0)  # [N, window_size]
        pad_mask = self.pad_mask[:, i : i + self.window_size].permute(
            1, 0
        )  # [N, window_size]
        return PaddedData(data=X, pad_mask=pad_mask)


class MultivariateTimeSeriesDataset(torch.utils.data.Dataset):
    """Time series dataset for use with Multivariate or Local models where given a
    a dataset containing multiple time series, N, we want to sample windows
    of a fixed length determined by the context_length and prediction_length
    across all time series so each sample will have shape (N, context_length + prediction_length)
    """

    def __init__(
        self,
        df: pd.DataFrame,
        context_length: int,
        prediction_length: int,
        conts: List[int] = [],
        lag_seq: List[int] = None,
        pad_fn=right_pad_sequence,
    ):
        self.df = df.copy()
        self.conts = conts.copy()

        self.lag_length = 0 if lag_seq is None else max(lag_seq)
        self.context_length = torch.tensor(context_length).long()
        self.prediction_length = torch.tensor(prediction_length).long()

        lengths = (
            self.df.groupby("unique_id", sort=False)["ds"].count()
            - context_length
            - prediction_length
            - self.lag_length
            + 1  # we return x and y so add 1 for teacher forcing
        ).clip(1)
        self.max_length = lengths.max()
        self.window_size = context_length + prediction_length + self.lag_length
        self.lengths = lengths
        self.pad_fn = pad_fn

    def build(self) -> "MultivariateTimeSeriesDataset":
        ts = []
        for unique_id, grp in self.df.groupby("unique_id", sort=False):
            ts.append(torch.from_numpy(grp[self.conts + ["y"]].values))
        ts, mask = self.pad_fn(ts, min_length=self.window_size)
        self.X = ts[:, :, :]
        self.pad_mask = mask
        return self

    def __len__(self) -> int:
        return self.max_length

    def __getitem__(self, i: int) -> PaddedData:
        X = self.X[0, i : i + self.window_size, :]
        pad_mask = self.pad_mask[0, i : i + self.window_size]
        return PaddedData(data=X, pad_mask=pad_mask)



import io
import zipfile
from datetime import datetime
from distutils.util import strtobool
from typing import Any, Callable, Tuple

import pandas as pd
import requests


# Converts the contents in a .tsf file into a dataframe and returns it along with other meta-data of the dataset: frequency, horizon, whether the dataset contains missing values and whether the series have equal lengths
#
# Parameters
# full_file_path_and_name - complete .tsf file path
# replace_missing_vals_with - a term to indicate the missing values in series in the returning dataframe
# value_column_name - Any name that is preferred to have as the name of the column containing series values in the returning dataframe
def _read(
    file,
    replace_missing_vals_with="NaN",
    value_column_name="series_value",
    encoding=None,
):
    col_names = []
    col_types = []
    all_data = {}
    line_count = 0
    frequency = None
    forecast_horizon = None
    contain_missing_values = None
    contain_equal_length = None
    found_data_tag = False
    found_data_section = False
    started_reading_data_section = False

    for line in file:
        if encoding is not None:
            line = line.decode(encoding)

        # Strip white space from start/end of line
        line = line.strip()

        if line:
            if line.startswith("@"):  # Read meta-data
                if not line.startswith("@data"):
                    line_content = line.split(" ")
                    if line.startswith("@attribute"):
                        if len(line_content) != 3:  # Attributes have both name and type
                            raise Exception("Invalid meta-data specification.")

                        col_names.append(line_content[1])
                        col_types.append(line_content[2])
                    else:
                        if len(line_content) != 2:  # Other meta-data have only values
                            raise Exception("Invalid meta-data specification.")

                        if line.startswith("@frequency"):
                            frequency = line_content[1]
                        elif line.startswith("@horizon"):
                            forecast_horizon = int(line_content[1])
                        elif line.startswith("@missing"):
                            contain_missing_values = bool(strtobool(line_content[1]))
                        elif line.startswith("@equallength"):
                            contain_equal_length = bool(strtobool(line_content[1]))

                else:
                    if len(col_names) == 0:
                        raise Exception(
                            "Missing attribute section. Attribute section must come before data."
                        )

                    found_data_tag = True
            elif not line.startswith("#"):
                if len(col_names) == 0:
                    raise Exception(
                        "Missing attribute section. Attribute section must come before data."
                    )
                elif not found_data_tag:
                    raise Exception("Missing @data tag.")
                else:
                    if not started_reading_data_section:
                        started_reading_data_section = True
                        found_data_section = True
                        all_series = []

                        for col in col_names:
                            all_data[col] = []

                    full_info = line.split(":")

                    if len(full_info) != (len(col_names) + 1):
                        raise Exception("Missing attributes/values in series.")

                    series = full_info[len(full_info) - 1]
                    series = series.split(",")

                    if len(series) == 0:
                        raise Exception(
                            "A given series should contains a set of comma separated numeric values. At least one numeric value should be there in a series. Missing values should be indicated with ? symbol"
                        )

                    numeric_series = []

                    for val in series:
                        if val == "?":
                            numeric_series.append(replace_missing_vals_with)
                        else:
                            numeric_series.append(float(val))

                    if numeric_series.count(replace_missing_vals_with) == len(
                        numeric_series
                    ):
                        raise Exception(
                            "All series values are missing. A given series should contains a set of comma separated numeric values. At least one numeric value should be there in a series."
                        )

                    all_series.append(pd.Series(numeric_series).array)

                    for i in range(len(col_names)):
                        att_val = None
                        if col_types[i] == "numeric":
                            att_val = int(full_info[i])
                        elif col_types[i] == "string":
                            att_val = str(full_info[i])
                        elif col_types[i] == "date":
                            att_val = datetime.strptime(
                                full_info[i], "%Y-%m-%d %H-%M-%S"
                            )
                        else:
                            raise Exception(
                                "Invalid attribute type."
                            )  # Currently, the code supports only numeric, string and date types. Extend this as required.

                        if att_val is None:
                            raise Exception("Invalid attribute value.")
                        else:
                            all_data[col_names[i]].append(att_val)

            line_count = line_count + 1

    if line_count == 0:
        raise Exception("Empty file.")
    if len(col_names) == 0:
        raise Exception("Missing attribute section.")
    if not found_data_section:
        raise Exception("Missing series information under data section.")

    all_data[value_column_name] = all_series

    return (
        all_data,
        frequency,
        forecast_horizon,
        contain_missing_values,
        contain_equal_length,
    )


# value_column_name - Any name that is preferred to have as the name of the column containing series values in the returning dataframe
def handle_zip_file_http_request(
    full_file_path_and_name: str,
    uri: str,
    reader: Callable[[io.BytesIO, str], Any],
):
    res = requests.get(uri)
    with zipfile.ZipFile(io.BytesIO(res.content)) as myzip:
        with myzip.open(full_file_path_and_name, mode="r") as file:
            return reader(file, encoding="cp1252")


def handle_filesystem_request(
    full_file_path_and_name: str,
    reader: Callable[[io.BytesIO, str], Any],
):
    with open(full_file_path_and_name, "rb") as file:
        return reader(file, encoding="cp1252")


def read(path: str, url: str = None) -> Tuple[Any, str, int, bool, bool]:
    if url is not None:
        return handle_zip_file_http_request(path, url, _read)
    else:
        return handle_filesystem_request(path, _read)



from typing import Dict, Optional

import torch
import torch.utils.data


def _calculate_seasonal_error(
    past_data: torch.Tensor,
    seasonality: Optional[int] = 1,
) -> torch.Tensor:
    """
    Calculates the seasonal error for a given time series data based on the specified seasonality.

    The seasonal error is computed by comparing the values of the time series with those
    lagged by the seasonality period. If the seasonality is larger than the length of the
    time series, the lag is set to 1 (t-1).

    Parameters
    ----------
    past_data: torch.Tensor
        A tensor containing the past time series data.
    seasonality: :obj:`int`
        The seasonality period. If not specified or if the length of past_data is less than the seasonality, a fallback value of 1 (i.e., t-1) is used.

    Returns
    -------
    torch.Tensor
        A tensor containing the mean absolute seasonal error.

    Example
    -------
    >>> past_data = torch.tensor([10.0, 20.0, 30.0, 40.0, 50.0])
    >>> seasonality = 2
    >>> error = _calculate_seasonal_error(past_data, seasonality)
    >>> print(error)
    tensor(20.0)
    """
    if seasonality < len(past_data):
        forecast_freq = seasonality
    else:
        forecast_freq = 1  # Fallback to t-1 if series length < seasonality. From Monash
        # https://github.com/rakshitha123/TSForecasting/blob/d187b057af25a203f8b321ad52edf11bc3c8b619/utils/error_calculator.R#L57

    y_t = past_data[:-forecast_freq]
    y_tm = past_data[forecast_freq:]

    result = torch.mean(torch.abs(y_t - y_tm))
    return result


def calculate_seasonal_error(
    trn_dl: torch.utils.data.DataLoader, seasonality: int
) -> torch.Tensor:
    seasonal_errors_per_series = []
    for series, mask in zip(trn_dl.dataset.X[:, :, 0], trn_dl.dataset.pad_mask):
        past_data = series[mask]
        error = _calculate_seasonal_error(past_data, seasonality)
        seasonal_errors_per_series.append(error)
    seasonal_errors = torch.stack(seasonal_errors_per_series, dim=0)
    return seasonal_errors.unsqueeze(1)


def mse(y_hat: torch.tensor, y: torch.tensor, dim=1) -> torch.tensor:
    """
    Calculates the Mean Squared Error (MSE) between the predicted values and the actual values.

    The function computes the MSE by taking the element-wise difference between the predicted
    values and the actual values, squaring these differences, and then averaging them. If the
    input tensors have more than one dimension, the averaging is done along the specified dimension.

    Parameters
    ----------
    y_hat: torch.tensor
        The predicted values.
    y: torch.tensor
        The actual values.
    dim: :obj: `int`, optional
        The dimension along which to compute the mean if the input tensors have more than one dimension. Defaults to 1.

    Returns
    -------
    torch.tensor
        A tensor containing the mean squared error.

    Example
    -------
    >>> y_hat = torch.tensor([2.5, 0.0, 2.0, 8.0])
    >>> y = torch.tensor([3.0, -0.5, 2.0, 7.0])
    >>> error = mse(y_hat, y)
    >>> print(error)
    tensor(0.3750)
    """
    return (
        ((y_hat - y) ** 2).mean(dim=dim)
        if len(y.shape) > 1
        else ((y_hat - y) ** 2).mean()
    )


def abs_error(y_hat: torch.tensor, y: torch.tensor, dim=1) -> torch.tensor:
    """
    Calculates the absolute error between the predicted values and the actual values.

    The function computes the absolute error by taking the element-wise absolute difference
    between the predicted values and the actual values, and then summing these differences.
    If the input tensors have more than one dimension, the summation is done along dimension 1.

    Parameters
    ----------
    y_hat: torch.tensor
        The predicted values.
    y: torch.tensor
        The actual values.

    Returns
    -------
    torch.tensor:
        A tensor containing the sum of absolute errors.

    Example
    -------
    >>> y_hat = torch.tensor([2.5, 0.0, 2.0, 8.0])
    >>> y = torch.tensor([3.0, -0.5, 2.0, 7.0])
    >>> error = abs_error(y_hat, y)
    >>> print(error)
    tensor(2.0)
    """
    return (
        (y_hat - y).abs().sum(dim=dim) if len(y.shape) > 1 else (y_hat - y).abs().sum()
    )


def abs_target_sum(y: torch.tensor, dim: int = 1) -> torch.tensor:
    """
    Calculates the sum of the absolute values of the target tensor.

    The function computes the sum of the absolute values of the target tensor. If the input
    tensor has more than one dimension, the summation is done along the specified dimension.

    Parameters
    ----------
    y: torch.tensor
        The target values.
    dim: :obj: `int`, optional
        The dimension along which to compute the sum if the input tensor has more than one dimension. Defaults to 1.

    Returns
    -------
    torch.tensor
        A tensor containing the sum of absolute values.

    Example
    -------
    >>> y = torch.tensor([3.0, -0.5, 2.0, -7.0])
    >>> result = abs_target_sum(y)
    >>> print(result)
    tensor(12.5)
    """
    return y.abs().sum(dim=dim) if len(y.shape) > 1 else y.abs().sum()


def abs_target_mean(y: torch.tensor, dim: int = 1) -> torch.tensor:
    """
    Calculates the mean of the absolute values of the target tensor.

    The function computes the mean of the absolute values of the target tensor. If the input
    tensor has more than one dimension, the mean is calculated along the specified dimension.

    Parameters
    ----------
    y: torch.tensor
        The target values.
    dim: :obj: `int`, optional
        The dimension along which to compute the mean if the input tensor has more than one dimension. Defaults to 1.

    Returns
    -------
    torch.tensor
        A tensor containing the mean of absolute values.

    Example
    -------
    >>> y = torch.tensor([[1.0, 2.0, 3.0, 4.0], [2.0, 4.0, 6.0, 8.0]])
    >>> result = abs_target_mean(y)
    >>> print(result)
    tensor([2.5, 5.0])
    """
    return y.abs().mean(dim=dim) if len(y.shape) > 1 else y.abs().mean()


def mae(y_hat: torch.tensor, y: torch.tensor, dim: int = 1) -> torch.tensor:
    """
    Calculates the Mean Absolute Error (MAE) between the predicted values and the actual values.

    The function computes the MAE by taking the element-wise absolute difference between the
    predicted values and the actual values, and then averaging these differences. If the input
    tensors have more than one dimension, the averaging is done along the specified dimension.

    Parameters
    ----------
    y_hat: torch.tensor)
        The predicted values.
    y: torch.tensor
        The actual values.
    dim: :obj:`int`, optional
        The dimension along which to compute the mean if the input tensors have more than one dimension. Defaults to 1.

    Returns
    -------
    torch.tensor
        A tensor containing the mean absolute error.

    Example
    -------
    >>> y_hat = torch.tensor([[1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0]])
    >>> y = torch.tensor([[2.0, 4.0, 6.0, 8.0], [1.0, 2.0, 3.0, 4.0]])
    >>> error = mae(y_hat, y)
    >>> print(error)
    tensor([2.5, 0.0])
    """
    return (
        (y_hat - y).abs().mean(dim=dim)
        if len(y.shape) > 1
        else (y_hat - y).abs().mean()
    )


def mase(
    y_hat: torch.tensor,
    y: torch.tensor,
    seasonal_errors: torch.tensor,
    dim: int = 1,
) -> torch.tensor:
    """
    Calculates the Mean Absolute Scaled Error (MASE) between the predicted values and the actual values.

    The function computes the MASE by dividing the mean absolute error (MAE) by the seasonal
    errors. If the input tensors have more than one dimension, the MAE is calculated along the
    specified dimension.

    Parameters
    ----------
    y_hat: torch.tensor)
        The predicted values.
    y: torch.tensor
        The actual values.
    seasonal_errors: torch.tensor)
        The seasonal errors used for scaling the mean absolute error (MAE).
    dim: :obj:`int`, optional
        The dimension along which to compute the mean if the input tensors have more than one dimension. Defaults to 1.

    Returns
    -------
    torch.tensor: A tensor containing the mean absolute scaled error.

    Example
    -------

    >>> y_hat = torch.tensor([1.0, 2.0, 3.0, 4.0])
    >>> y = torch.tensor([0.0, 4.0, 6.0, 8.0])
    >>> seasonal_errors = torch.tensor([2.0])
    >>> error = mase(y_hat, y, seasonal_errors)
    >>> print(error)
    tensor([1.25])
    """
    if len(y.shape) > 2 and len(seasonal_errors.shape) == 1:
        seasonal_errors = seasonal_errors.unsqueeze(-1)
    return mae(y_hat, y, dim=dim) / seasonal_errors


def rmsse(
    y_hat: torch.tensor,
    y: torch.tensor,
    seasonal_errors: torch.tensor,
    dim: int = 1,
) -> torch.tensor:
    """
    Calculates the Mean Absolute Scaled Error (MASE) between the predicted values and the actual values.

    The function computes the MASE by dividing the mean absolute error (MAE) by the seasonal
    errors. If the input tensors have more than one dimension, the MAE is calculated along the
    specified dimension.

    Parameters
    ----------
    y_hat: torch.tensor)
        The predicted values.
    y: torch.tensor
        The actual values.
    seasonal_errors: torch.tensor)
        The seasonal errors used for scaling the mean absolute error (MAE).
    dim: :obj:`int`, optional
        The dimension along which to compute the mean if the input tensors have more than one dimension. Defaults to 1.

    Returns
    -------
    torch.tensor: A tensor containing the mean absolute scaled error.

    Example
    -------

    >>> y_hat = torch.tensor([1.0, 2.0, 3.0, 4.0])
    >>> y = torch.tensor([0.0, 4.0, 6.0, 8.0])
    >>> seasonal_errors = torch.tensor([2.0])
    >>> error = mase(y_hat, y, seasonal_errors)
    >>> print(error)
    tensor([1.25])
    """
    if len(y.shape) > 2 and len(seasonal_errors.shape) == 1:
        seasonal_errors = seasonal_errors.unsqueeze(-1)
    return mse(y_hat, y, dim=dim).sqrt() / seasonal_errors


def _single_ts_mape(y_hat: torch.tensor, y: torch.tensor) -> torch.tensor:
    ape = (y_hat - y).abs() / y.abs()
    mask = torch.isfinite(ape)
    ape_filtered = ape[mask]
    result = torch.mean(ape_filtered)
    return result


def mape(y_hat: torch.tensor, y: torch.tensor) -> torch.tensor:
    """
    Calculates the Mean Absolute Percentage Error (MAPE) between the predicted values and the actual values.

    The function computes the MAPE by calculating the absolute percentage error for each time
    series and then averaging these errors. If the input tensors contain multiple time series
    (i.e., are two-dimensional), the MAPE is calculated for each time series independently.

    Parameters
    ----------
    y_hat: torch.tensor
        The predicted values.
    y: torch.tensor)
        The actual values.

    Returns
    -------
    torch.tensor: A tensor containing the mean absolute percentage error for each time series.

    Example
    -------
    >>> y_hat = torch.tensor([1.0, 2.0, 3.0, 4.0])
    >>> y = torch.tensor([1.0, 1.0, 1.0, 1.0])
    >>> error = mape(y_hat, y)
    >>> print(error)
    tensor([1.5])

    >>> y_hat = torch.tensor([[1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0]])
    >>> y = torch.tensor([[0.0, 4.0, 6.0, 8.0], [0.0, 2.0, 3.0, 4.0]])
    >>> error = mape(y_hat, y)
    >>> print(error)
    tensor([0.5, 0.0])

    """
    if len(y.shape) > 1:
        return torch.stack(
            [_single_ts_mape(y_hat[i], y[i]) for i in range(y.shape[0])], dim=0
        )
    else:
        return _single_ts_mape(y_hat, y)


def smape(y_hat: torch.tensor, y: torch.tensor, dim: int = 1) -> torch.tensor:
    """
    Calculates the Symmetric Mean Absolute Percentage Error (sMAPE) between the predicted values and the actual values.

    The function computes the sMAPE by calculating the symmetric absolute percentage error
    for each prediction, which is given by the formula 2 * |y_hat - y| / (|y| + |y_hat|).
    If the input tensors have more than one dimension, the mean is calculated along the specified dimension.

    Parameters
    ----------
    y_hat: torch.tensor
        The predicted values.
    y: torch.tensor
        The actual values.
    dim: :obj:`int`, optional
        The dimension along which to compute the mean if the input tensors have more than one dimension. Defaults to 1.

    Returns
    -------
    torch.tensor
        A tensor containing the symmetric mean absolute percentage error.

    Example
    -------
    >>> y_hat = torch.tensor([[1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0]])
    >>> y = torch.tensor([[0.0, 4.0, 6.0, 8.0], [0.0, 2.0, 3.0, 4.0]])
    >>> error = smape(y_hat, y)
    >>> print(error)
    tensor([1.0, 0.5])

    """
    pe = 2 * (y_hat - y).abs() / (y.abs() + y_hat.abs())
    return pe.mean(dim=dim) if len(y.shape) > 1 else pe.mean()


def msmape(y_hat: torch.tensor, y: torch.tensor, dim: int = 1, eps=0.1) -> torch.tensor:
    """
    Calculates the Modified Symmetric Mean Absolute Percentage Error (msMAPE) between the predicted values and the actual values.

    The function computes the msMAPE by calculating the modified symmetric absolute percentage error
    for each prediction, which is given by the formula 2 * |y_hat - y| / max((|y| + |y_hat| + eps), (0.5 + eps)).
    If the input tensors have more than one dimension, the mean is calculated along the specified dimension.

    Parameters
    ----------
    y_hat: torch.tensor)
        The predicted values.
    y: torch.tensor)
        The actual values.
    dim: :obj:`int`, optional
        The dimension along which to compute the mean if the input tensors have more than one dimension. Defaults to 1.
    eps: :obj:`float`, optional
        Epsilon value to prevent division by zero. Defaults to 0.1.

    Returns
    -------
    torch.tensor
        A tensor containing the modified symmetric mean absolute percentage error.


    Example
    -------
    >>> y_hat = torch.tensor([[1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0]])
    >>> y = torch.tensor([[0.0, 4.0, 6.0, 8.0], [0.0, 2.0, 3.0, 4.0]])
    >>> error = msmape(y_hat, y)
    >>> print(error)
    tensor([1.0, 0.5])

    """
    pe = (
        2
        * (y_hat - y).abs()
        / torch.max((y.abs() + y_hat.abs() + eps), torch.ones_like(y) * 0.5 + eps)
    )
    return pe.mean(dim=dim) if len(y.shape) > 1 else pe.mean()


def nd(y_hat: torch.tensor, y: torch.tensor, dim: int = 1) -> torch.tensor:
    return abs_error(y_hat, y, dim=dim) / abs_target_sum(y, dim=dim)


def get_metrics_per_ts(
    y_hat: torch.tensor, y: torch.tensor, seasonal_error: torch.tensor
) -> Dict[str, torch.tensor]:
    mase_per_ts = mase(y_hat, y, seasonal_error)
    mase_per_ts = mase_per_ts[mase_per_ts.isfinite()]
    rmsse_per_ts = rmsse(y_hat, y, seasonal_error)
    rmsse_per_ts = rmsse_per_ts[rmsse_per_ts.isfinite()]
    return {
        "mse": mse(y_hat, y),
        "abs_error": abs_error(y_hat, y),
        "abs_target_sum": abs_target_sum(y),
        "abs_target_mean": abs_target_mean(y),
        "mase": mase_per_ts,
        "rmsse": rmsse_per_ts,
        "mape": mape(y_hat, y),
        "smape": smape(y_hat, y),
        "msmape": msmape(y_hat, y),
        "nd": nd(y_hat, y),
        "mae": mae(y_hat, y),
        "rmse": mse(y_hat, y).sqrt(),
        "seasonal_error": seasonal_error,
    }


def aggregate_base_metrics(metrics_per_ts: Dict[str, torch.tensor]) -> Dict[str, float]:
    aggregate_map = {
        "mse": "mean",
        "abs_error": "sum",
        "abs_target_sum": "sum",
        "abs_target_mean": "mean",
        "seasonal_error": "mean",
    }
    return {
        k: getattr(v, aggregate_map[k])().item()
        for k, v in metrics_per_ts.items()
        if k in aggregate_map
    }


def aggregate_mean_metrics(metrics_per_ts: Dict[str, torch.tensor]) -> Dict[str, float]:
    aggregate_map = {
        "mase": "mean",
        "rmsse": "mean",
        "mape": "nanmean",
        "smape": "mean",
        "msmape": "mean",
        "mae": "mean",
        "rmse": "mean",
    }
    return {
        f"mean_{k}": getattr(v, aggregate_map[k])().item()
        for k, v in metrics_per_ts.items()
        if k in aggregate_map
    }


def aggregate_median_metrics(
    metrics_per_ts: Dict[str, torch.tensor],
) -> Dict[str, float]:
    aggregate_map = {
        "mase": "median",
        "rmsse": "median",
        # "mape": "nanmedian",
        "smape": "median",
        "msmape": "median",
        "mae": "median",
        "rmse": "median",
    }
    return {
        f"median_{k}": getattr(v, aggregate_map[k])().item()
        for k, v in metrics_per_ts.items()
        if k in aggregate_map
    }


def calc_metrics(
    y_hat: torch.tensor, y: torch.tensor, seasonal_error: torch.tensor
) -> Dict[str, float]:
    metrics_per_ts = get_metrics_per_ts(y_hat, y, seasonal_error)
    base_metrics = aggregate_base_metrics(metrics_per_ts)
    mean_metrics = aggregate_mean_metrics(metrics_per_ts)
    median_metrics = aggregate_median_metrics(metrics_per_ts)
    metrics = {**base_metrics, **mean_metrics, **median_metrics}
    return metrics
